/*
** cscan.l
**
** Does the major work of removing identifiers and constants
** from the input stream, for Cxref. Its output is then extensively
** postprocessed.
**
** Arnold Robbins, Information and Computer Science, Georgia Tech
**	gatech!arnold
** Copyright (c) 1984 by Arnold Robbins.
** All rights reserved.
** This program may not be sold, but may be distributed
** provided this header is included.
*/

%{
extern int line_no;
extern char *fname, *basename();
%}

letter		[A-Za-z_$]
digit		[0-9]
hexdigit	[0-9A-Fa-f]
integer_suffix	([lL]?[uU]?|[uU]?[lL]?)
floating_suffix	[fFlL]?
trigraph	[?][?][=(/)'<!>>-]
backslash	(\\|[?][?]\/)

%%
int		|
char		|
float		|
double		|
struct		|
union		|
long		|
short		|
unsigned	|
auto		|
extern		|
register	|
typedef		|
static		|
goto		|
return		|
sizeof		|
break		|
continue       	|
if		|
else		|
for		|
do		|
while		|
switch		|
case		|
default		|
entry		|
enum		|
void		|
const		|
volatile	|
noalias		|
signed		|
near		|
far		|
huge		|
cdecl		|
fortran		|
pascal		|
interrupt	|
inline		|
define		|
undef		|
include		|
ifdef		|
ifndef		|
elif		|
defined		|
pragma		|
endif		;	/* ignore C and cpp keywords */

{trigraph}	;	/* ignore trigraphs */


"<".+">"	outid();	/* DON'T forget about include-file names */

"\n"		line_no++;

"/*"		{	/* get rid of comments */
			register char c, c1;

		  loop: while((c = input()) != '*' && c != 0)
				if(c == '\n')
					line_no++;

			if(c == 0)
			{
				fprintf(stderr,
				"unexpected EOF in comment at line %d, file %s\n",
					line_no, basename(fname));
				exit(1);
			}

			if((c1 = input()) != '/')
			{
				unput(c1);	/* could be '*' before '/' */
				goto loop;
			}
		}

{letter}({letter}|{digit})*	outid();  /* what we actually want */


L?'[^\\']'		|
L?'{backslash}x{hexdigit}+'	|
L?'{backslash}{digit}{1,3}'	|
L?'{backslash}{backslash}'	|
L?'{backslash}[\\abfnrtv"?']'		outchar();

L?\"			{	/* collect quoted strings */
				register char c;
				register int i;

				for(i = (yytext[0] == 'L') ? 2 : 1,
					c = input(); ; i++, c = input())
					switch (c) {
					case '"':
						yytext[i] = c;
						yytext[++i] = '\0';
						yyleng = i - 1;
						goto fini;

					case '\\':
						yytext[i] = '\\';
						yytext[++i] = input();
						if (yytext[i] == '\n')
						{
							line_no++;
							yytext[i] = 'N';
							/* make visible */
						}
						break;

					case '?': /* handle trigraphs */
						yytext[i] = '?';
						yytext[++i] = input();
						if (yytext[i] != '?')
						{	/* not trigraph */
							unput(yytext[i--]);
							break;
						}
						yytext[++i] = input();
						if (yytext[i] == '/')
							yytext[++i] = input();
						break;

					case 0:
						fprintf(stderr,
					"unexpected EOF inside string at line %d, file %s\n",
							line_no, basename(fname));
						exit(1);
						break;

					default:
						yytext[i] = c;
						break;
					}

				fini:
					outstring();
		}

[+-]?{digit}+{integer_suffix}			|
[+-]?0[Xx]({digit}|[a-fA-F])+{integer_suffix}		outint();

[+-]?{digit}*"."{digit}+([Ee][+-]?{digit}+)?{floating_suffix}	|
[+-]?{digit}+"."{digit}*([Ee][+-]?{digit}+)?{floating_suffix}	|
[+-]?{digit}+[Ee][+-]?{digit}+{floating_suffix}		outfloat();

.			;	/* delete everything else */

%%

yywrap()	/* wrap up for lex, return 1 */
{
	return(1);
}

#include "constdefs.h"

extern char *fname;
extern char *basename();

outchar()
{
	outtext(CHAR);
}

outstring()
{
	outtext(STRING);
}

outint()
{
	int i = strlen(yytext);

	/* handle long integer constants */

#if 0
	/* NO: suffixes should be preserved! */
	if (yytext[i-1] == 'l' || yytext[i-1] == 'L')
		yytext[i-1] = '\0';
#endif

	outtext(INT);
}

outfloat()
{
	outtext(FLOAT);
}

outtext(type)
char type;
{
	printf("~%c%s\t%s\t%d\n", type, yytext, basename(fname), line_no);
}
